#   Copyright and License: see LICENSE file
#
#   Contributors:
#	   * Joachim Baran

class Session
  @@db_handles = {}
  @@gazebo_handle = nil
  @@salts = nil

  def initialize(db_type, session)
    @db_type = db_type
    @session = session

    Synchronization.instance.lock('session:salts')
    unless @@salts then
      if USE_GDBM then
        # Evaluate whether it is necessary to use GDBM::NOLOCK.
        @@salts = GDBM.open(SALTS, 0660, GDBM::WRCREAT | GDBM::SYNC)
      else
        @@salts = SDBM.open(SALTS, 0660)
      end
    end
    Synchronization.instance.unlock('session:salts')
  end

  def connect(options, database, username, password, hostname="localhost", port="5432")
    @db_database = database
    @db_username = 'gazebo'
    @db_password = 'gazebo'
    @db_hostname = hostname
    @db_port = port
    @restricted = false
    
    unless options['bypass_gazebo'] then
      # Connect to Gazebo's database, which contains user-, group- and
      # meta-data about the whole system.
      begin
        Synchronization.instance.lock('session:gazebo')
        unless @@gazebo_handle then
          @@gazebo_handle = DBI.connect("DBI:#{@db_type}:database=gazebo;host=#{@db_hostname};port=#{@db_port}", @db_username, @db_password)
        end
        Synchronization.instance.unlock('session:gazebo')
      rescue DBI::OperationalError
        Synchronization.instance.unlock('session:gazebo')
        raise SessionError.new('Cannot connect to Gazebo database.')
      end

      # Try to retrieve session status from Gazebo's database and
      # if this fails due to a DB error, then try to create the Gazebo
      # tables.
      retried = false
      valid_session = nil
      begin
        valid_session = is_valid_session()
      rescue DBI::Error
        create_gazebo_tables()
        if retried == false then
          retried = true
          retry
        else
          raise SessionError.new('Unable to create Gazebo tables from scratch.')
        end
      end

      # TODO Should this come after username/password validation?
      if options['restricted'] then
        @restricted = true
        return nil
      end

      # Evaluate username and password and generate a session ID.
      unless valid_session or verify_password(username, password) then
        raise GazeboAuthenticationError.new("Wrong username and/or password?")
      end

      @session = generate_session(username) unless is_valid_session()
    end

    return @session unless @db_database

    # Connect to the database where the actual data is stored.
    begin
      Synchronization.instance.lock('session:db')

      connection_parameters = "DBI:#{@db_type}:database=#{@db_database};host=#{@db_hostname};port=#{@db_port}"
      db_key = "#{connection_parameters};#{@db_username};#{Digest::SHA1::hexdigest(@db_password)}"

      Log.instance.log('DB KEY: ' << db_key)
      if @@db_handles.has_key?(db_key) then
        @db_handle = @@db_handles[db_key]
      else
        @db_handle = DBI.connect(connection_parameters, @db_username, @db_password)
        get_databases() # Sanity check to see whether we have implemented methods
                        # for this particular driver. If not, we get a NoMethodError.
        @@db_handles[db_key] = @db_handle
      end

      Synchronization.instance.unlock('session:db')
    rescue DBI::InterfaceError
      Synchronization.instance.unlock('session:db')
      raise SessionError.new("Could not load driver: #{@db_type}")
    rescue NoMethodError
      Synchronization.instance.unlock('session:db')
      raise SessionError.new()
    rescue DBI::OperationalError
      Synchronization.instance.unlock('session:db')
      raise SessionError.new("Could not connect to database: #{@db_database}")
    end

    return @session
  end

  def disconnect()
    simple_delete('gazebo_sessions', 'session_id = ?', @session)
  end

  def log(message)
    # TODO
  end

  def create_gazebo_tables()
    self.send(@db_type.downcase + '__create_gazebo_tables')
  end

  def get_databases()
    return self.send(@db_type.downcase + '__get_databases')
  end

  def get_tables()
    return self.send(@db_type.downcase + '__get_tables')
  end

  def query(options, result_set_columns, table_aliases, query, arguments)
    unless options['bypass_gazebo'] then
      raise SessionError("Please renew session.") unless is_valid_session()
    end
    self.send(@db_type.downcase + '__query', options, result_set_columns, table_aliases, query, arguments)
  end

  def query_union(options, result_set_columns, table_aliases, queries, arguments)
    unless options['bypass_gazebo'] then
      raise SessionError("Please renew session.") unless is_valid_session()
    end
    self.send(@db_type.downcase + '__query_union', options, result_set_columns, table_aliases, queries, arguments)
  end

  def update(options, table_alias, update_columns, update_arguments, query, arguments)
    unless options['bypass_gazebo'] then
      raise SessionError("Please renew session.") unless is_valid_session()
    end
    self.send(@db_type.downcase + '__update', options, table_alias, update_columns, update_arguments, query, arguments)
  end

  def create_data(options, table_alias, columns, arguments)
    retries = 30 # Arbitrarily chosen.
    Log.instance.log('Prepare data..')
    begin
      limbo_key = get_random_hash()
      Log.instance.log('Limbo key: ' << limbo_key)
      self.send(@db_type.downcase + '__insert',
        options,
        table_alias,
        [ 'limbo_key' ] + columns,
        [ limbo_key ] + arguments
      )
    rescue DBI::ProgrammingError => e
      if retries > 0 then
        retries -= 1
        retry
      else
	Log.instance.log(e.message)
        # If you get this exception, check whether you have run 'create_stocks.rb'!
        raise SessionError.new('Could not generate a unique limbo key.')
      end
    end
    Log.instance.log('Data in limbo..')

    data_id_set = query({}, [ 'id' ], [ table_alias ], 'limbo_key=?', [ limbo_key ])
    raise SessionError.new('Lost data in limbo. Woops.') unless data_id_set and data_id_set.size == 1

    Log.instance.log('Returning: ' << data_id_set[0][0])
    return data_id_set[0][0]
  end

  def update_data(options, data_id, table_alias, update_columns, update_arguments)
    Log.instance.log('Update data.. ID: ' << data_id)
    # TODO Deal with data in limbo.
    update_columns << 'limbo_key' << 'created_by' << 'created_on'
    update_arguments << nil << get_username_by_session() << Time.now.strftime("%Y-%m-%d %H:%M:%S")
    update(options, table_alias, update_columns, update_arguments, 'id=?', [ data_id ])
    Log.instance.log('Updated!')
  end

  def query_data(options, result_set_columns, table_aliases, query, arguments)
    # TODO Figure out in which groups a datum is...
    query = 'limbo_key IS NULL AND ' << query
    query(options, result_set_columns, table_aliases, query, arguments)
  end

  def create_internals(options, table_alias, columns, arguments)
    # TODO Check permissions.
    if options['generate_password'] then
      pwd = get_random_password()
      columns << 'pwd_hash'
      arguments << hash_password(get_username_by_session(), Digest::SHA1::hexdigest(pwd))
    end

    columns << 'created_by' << 'created_on'
    arguments << get_username_by_session() << Time.now.strftime("%Y-%m-%d %H:%M:%S")
    Log.instance.log('INSERT: ' << arguments.join(':'))
    Log.instance.log('ABCDEF: ' << columns.join(':'))
    simple_insert(table_alias, arguments, columns)

    # Only send the email when the user was successfully added to the database.
    email_to = options['email_to']
    if email_to and pwd then
      Log.instance.log('SENDING MAIL...')
      Net::SMTP.start(GAZEBO_EMAIL_HOST,
        GAZEBO_EMAIL_PORT,
        GAZEBO_EMAIL_DOMAIN,
        GAZEBO_EMAIL_ADDRESS,
        GAZEBO_EMAIL_PASSWORD,
        :login
      ) { |smtp|
        smtp.open_message_stream(GAZEBO_EMAIL_ADDRESS, [ email_to ]) { |mail|
          mail.puts 'From: ' << GAZEBO_EMAIL_ADDRESS
          mail.puts 'To: ' << email_to
          mail.puts 'Subject: [' << MONIKER << '] Your Password'
          mail.puts 'Hello new ' << MONIKER << '-user,'
          mail.puts ''
          mail.puts 'An account has been created for you that grants you access to ' << MONIKER << '.'
          mail.puts ''
          mail.puts 'Please use your browser to log-in on ' << GAZEBO_HOST
          mail.puts ''
          mail.puts 'Your usename is: ' << arguments[0]
          mail.puts 'Your password is: ' << pwd
          mail.puts ''
          mail.puts 'To change your password click on \"Administration\" once you have logged in.'
          mail.puts ''
          mail.puts 'Toodle-oo,'
          mail.puts MONIKER
        }
      }
      Log.instance.log('SENT...')
    end

    return true
  end

  def delete_internals(options, table_alias, query, arguments)
    # TODO Check permissions.
    @db_handle = @@gazebo_handle
    simple_delete(table_alias, query, arguments)
    return 0
  end

  def update_internals(options, table_alias, key, key_column, columns, arguments)
    # TODO Check permissions.
    @db_handle = @@gazebo_handle
    update(options, table_alias, columns, arguments, "#{key_column}=?", [ key ])
    Log.instance.log('Updated!')
    return 0
  end

  def deactivate_user(options, username)
    # TODO
    update_internals(options, 'gazebo_users', username, 'username', [ 'deactivated' ], [ true ])
  end

  def update_subscription(options, name, username, is_admin, is_subscribed)
    # TODO Check permissions.
    @db_handle = @@gazebo_handle

    # Get the group ID first, which is needed a couple of times below:
    result = simple_query([ 'group_id' ], [ 'gazebo_groups' ], 'name = ?', [ name ])
    group_id = result[0][0]

    # See whether a subscription exists already:
    result = simple_query([ 'group_id' ],
      [ 'gazebo_subscriptions' ],
      'group_id = ? AND username = ?',
      [ group_id, username ])

    case result.length
    when 1 then
      if is_subscribed then
        update(options, 'gazebo_subscriptions', [ 'group_admin' ], [ is_admin ],
          'group_id = ? AND username = ?', [ group_id, username ])
      else
        simple_delete('gazebo_subscriptions', 'group_id = ? AND username = ?', [ group_id, username ])
      end
    when 0 then
      create_internals(options, 'gazebo_subscriptions',
        [ 'group_id', 'username', 'group_admin' ], [ group_id, username, is_admin ])
    else
      # Woops...
    end
  end

  # Does not raise exception when not logged in, but tells you so.
  def get_status()
    username = get_username_by_session()
    return { 'logged_in' => username != nil, 'username' => username }
  end

  #private

  # Generic implementation for accessing Gazebo's tables:

  def get_random_password()
    alphabet = (0..9).to_a + ('a'..'z').to_a + ('A'..'Z').to_a
    return Array.new(8+rand(4)).map!{ alphabet[rand(alphabet.length)] }.join()
  end

  def get_random_hash()
    seed = Array.new(2048).map!{ (32 + rand(94)).chr }.join()
    return Digest::SHA1::hexdigest(seed)
  end

  def get_random_salt()
    alphabet = (0..9).to_a + ('a'..'z').to_a + ('A'..'Z').to_a +
      [ '!', '@', '$', '%', '^', '&', '*', '(', ')', '_', '-', '=', '+' ] +
      [ '[', ']', '{', '}', ':', ';', '|', '\\', '<', '>', ',', '.', '/', '?', '~' ]
    return Array.new(16+rand(8)).map!{ alphabet[rand(alphabet.length)] }.join()
  end

  def get_list(columns, tables, join_table, options, query, arguments)
    unless options['bypass_gazebo'] then
      raise SessionError("Please renew session.") unless is_valid_session()
    end
    query << ' ORDER BY ' << options['orderby'] if options['orderby']
    return simple_query(columns, tables, query, arguments) unless options['comprehensive']
    return simple_left_join_query(columns, tables, [ join_table ], query, arguments)
  end

  def is_valid_session()
    return get_username_by_session() != nil
  end

  def get_username_by_session()
    return nil unless @session
    result = simple_query([ 'username' ], [ 'gazebo_sessions' ], 'session_id = ?', @session)
    return nil if result.size != 1
    return result[0][0]
  end

  def verify_password(username, purported_password)
    return nil unless username or purported_password

    result = simple_query([ 'pwd_hash' ], [ 'gazebo_users' ], 'username = ?', username)
    return nil if result.size != 1

    Synchronization.instance.lock('session:salts')
    salt = @@salts[username]
    Synchronization.instance.unlock('session:salts')

    return nil unless salt
    pwd_hash = Digest::SHA1::hexdigest(purported_password + salt)
    return result[0][0] == pwd_hash
  end

  def hash_password(username, password)
    salt = get_random_salt()

    Synchronization.instance.lock('session:salts')
    @@salts[username] = salt
    Synchronization.instance.unlock('session:salts')

    return Digest::SHA1::hexdigest(password + salt)
  end

  def generate_session(username)
    simple_delete('gazebo_sessions', 'username = ?', username)
    retries = 30 # Arbitrarily chosen.
    begin
      session_id = get_random_hash()
      simple_insert('gazebo_sessions', [ session_id, username ], nil)
    rescue DBI::ProgrammingError
      if retries > 0 then
        retries -= 1
        retry
      else
        raise SessionError.new('Could not generate a unique session ID.')
      end
    end
    return session_id
  end

  # TODO Merge simple_query and simple_left_join_query?
  def simple_query(result_set_columns, table_aliases, query, arguments)
    sql_query = 'SELECT ' <<
        result_set_columns * ',' <<
        ' FROM ' <<
        table_aliases * ',' <<
        ' WHERE ' <<
        query

    statement = @@gazebo_handle.prepare(sql_query)
    statement.execute(*arguments)
    result = statement.fetch_all
    statement.finish

    return result
  end

  def simple_left_join_query(result_set_columns, table_aliases, join_table_aliases, query, arguments)
    sql_query = 'SELECT ' <<
        result_set_columns * ',' <<
        ' FROM ' <<
        table_aliases * ',' <<
        ' LEFT JOIN ' <<
        join_table_aliases * ',' <<
        ' ON ' <<
        query

    statement = @@gazebo_handle.prepare(sql_query)
    statement.execute(*arguments)
    result = statement.fetch_all
    statement.finish

    return result
  end

  def simple_insert(table_alias, arguments, optional_columns)
    sql_insert = 'INSERT INTO ' << table_alias
    sql_insert << ' (' << optional_columns.join(',') << ') ' if optional_columns
    sql_insert << ' VALUES (' <<
      ('?' * arguments.size).scan(/./).join(',') <<
      ')'

    statement = @@gazebo_handle.prepare(sql_insert)
    statement.execute(*arguments)
    statement.finish
  end

  def simple_delete(table_alias, condition, arguments)
    sql_delete = 'DELETE FROM ' <<
      table_alias <<
      ' WHERE ' <<
      condition

    statement = @@gazebo_handle.prepare(sql_delete)
    statement.execute(*arguments)
    statement.finish
  end

  private

  # PostgreSQL implementation:

  def pg__create_gazebo_tables()
    handle = DBI.connect("DBI:#{@db_type}:database=gazebo;host=#{@db_hostname};port=#{@db_port}", @db_username, @db_password)

    handle.execute('CREATE TABLE gazebo_context (' <<
      'version varchar(50)' <<
      ')')

    handle.execute('CREATE TABLE gazebo_log (' <<
      'log_id SERIAL PRIMARY KEY,' <<
      'occurred_on timestamp,' <<
      'version varchar(50),' <<
      'username varchar(40),' <<
      'class varchar(20),' <<
      'message text' <<
      ')')

    handle.execute('CREATE TABLE gazebo_users (' <<
      'username varchar(40) PRIMARY KEY,' <<
      'created_by varchar(40),' <<
      'created_on timestamp,' <<
      'pwd_hash varchar(128),' << # Enough room for 1024-bit digests.
      'firstname varchar(40),' <<
      'lastname varchar(40),' <<
      'email varchar(256),' <<
      'deactivated boolean,' <<
      'create_user boolean,' <<
      'create_group boolean,' <<
      'permissions_user int' <<
      ')')

    # only applies to owner if private == true
    # only applies to owner and group members if private == false and public == false
    # applies to everyone if private == false and public == true
    handle.execute('CREATE TABLE gazebo_groups (' <<
      'group_id SERIAL PRIMARY KEY,' <<
      'created_by varchar(40),' <<
      'created_on timestamp,' <<
      'contact varchar(40),' <<
      'name varchar(200),' <<
      'description text,' <<
      'contribute_permissions int,' <<
      'visible_permissions int,' <<
      'modify_permissions int' <<
      ')')

    handle.execute('CREATE TABLE gazebo_subscriptions (' <<
      'username varchar(40),' <<
      'created_by varchar(40),' <<
      'created_on timestamp,' <<
      'group_id int,' <<
      'group_admin boolean' <<
      ')')

    handle.execute('CREATE TABLE gazebo_metadata (' <<
      'data_id bigint,' <<
      'created_by varchar(40),' <<
      'created_on timestamp,' <<
      'group_id int' <<
      ')')

    handle.execute('CREATE TABLE gazebo_sessions (' <<
      'session_id varchar(128) PRIMARY KEY,' << # session_id is an up to 1024-bit hash
      'username varchar(40)' <<
      ')')

    # Set-up administrator account and public group
    now = Time.now

    admin_username = 'Administrator'
    pwd_hash = hash_password(admin_username, Digest::SHA1::hexdigest('ChangeMeImmediatelyPlease'))

    # Password: 'ChangeMeImmediatelyPlease' without the quotes
    handle.execute('INSERT INTO gazebo_users VALUES (' <<
      '\'' << admin_username << '\',' <<
      '\'' << admin_username << '\',' <<
      '\'' << now.strftime("%Y-%m-%d %H:%M:%S") << '\',' <<
      '\'' << pwd_hash << '\',' <<
      '\'\',' <<
      '\'\',' <<
      '\'\',' <<
      'false,' <<
      'true,' <<
      'true,' <<
      '2' <<
      ')')

    handle.execute('INSERT INTO gazebo_groups ' <<
      '(created_by, created_on, contact, name, description, contribute_permissions, visible_permissions, modify_permissions)' <<
      ' VALUES (' <<
      '\'' << admin_username << '\',' <<
      '\'' << now.strftime("%Y-%m-%d %H:%M:%S") << '\',' <<
      '\'' << admin_username << '\',' <<
      '\'Public\',' <<
      '\'Publicly available data.\',' <<
      '2,' <<
      '2,' <<
      '0' <<
      ')')

    handle.execute('INSERT INTO gazebo_subscriptions VALUES (' <<
      '\'' << admin_username << '\',' <<
      '\'' << admin_username << '\',' <<
      '\'' << now.strftime("%Y-%m-%d %H:%M:%S") << '\',' <<
      '1,' <<
      'true' <<
      ')')

    # Create example dataset for SuggestionTextField:
    #handle.execute('CREATE TABLE gazebo_example_suggestiontextfield_overview (' <<
    #  'species varchar(100),' <<
    #  'occurrences integer' <<
    #  ')')
    #handle.execute('CREATE TABLE gazebo_example_suggestiontextfield_refinement (' <<
    #  'abstract varchar(100),' <<
    #  'refined varchar(100)' <<
    #  ')')

    # TODO Add data here...

    handle.execute('INSERT INTO gazebo_context VALUES (' <<
      '\'' << GAZEBO_VERSION << '\'' <<
      ')')

    log('gazebo:created')

    handle.disconnect()
  end

  def pg__get_databases()
    result = @db_handle.execute('SELECT d.datname FROM pg_catalog.pg_database d')
    return result.collect { |row| row.to_s }
  end

  def pg__get_tables()
    result = @db_handle.execute('SELECT c.relname FROM pg_catalog.pg_class c LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace WHERE c.relkind = \'r\' AND pg_catalog.pg_table_is_visible(c.oid)')
    return result.collect { |row| row.to_s }
  end

  # Eventually, query will become a Tree::TreeNode. Or maybe not.
  def pg__query(options, result_set_columns, table_aliases, query, arguments)
    sql_query = 'SELECT ' <<
        result_set_columns * ',' <<
        ' FROM ' <<
        table_aliases * ',' <<
        ' WHERE ' <<
        query

    if options['limit'] then
      sql_query << ' LIMIT ' << options['limit'].to_s
    end

    Log.instance.log('RAW SQL QUERY: ' << sql_query)
    Log.instance.log('RAW ARGUMENTS: ' << arguments.to_s)
    statement = @db_handle.prepare(sql_query)
    statement.execute(*arguments)
    result = statement.fetch_all
    statement.finish
    Log.instance.log('RAW RESULTSET: ' << result.to_s)
    Log.instance.log('RAW ISNULL   : ' << (result == nil).to_s)

    if options['count'] then
      count_query = 'SELECT ' <<
        'COUNT(' << result_set_columns * ',' << ')' <<
        ' FROM ' <<
        table_aliases * ',' <<
        ' WHERE ' <<
        query
      statement = @db_handle.prepare(count_query)
      statement.execute(*arguments)
      count = statement.fetch_all
      statement.finish
      result = [ count, result ]
    end
    Log.instance.log('COUNT DONE')

    return result
  end

  def pg__query_union(options, result_set_columns, table_aliases, queries, arguments)
    sql_query = []
    result_set_columns.each_index { |index|
      sql_query << ('SELECT ' <<
          result_set_columns[index] * ',' <<
          ' FROM ' <<
          table_aliases[index] * ',' <<
          ' WHERE ' <<
          queries[index])
    }
    sql_query = sql_query.join(' UNION ')

    if options['limit'] then
      sql_query << ' LIMIT ' << options['limit'].to_s
    end

    statement = @db_handle.prepare(sql_query)
    statement.execute(*arguments)
    result = statement.fetch_all
    statement.finish

    return result
  end

  def pg__insert(options, table_alias, columns, arguments)
    sql_columns = ''
    sql_columns = '(' << columns.join(',') << ')' if columns and columns.size > 0
    sql_insert = 'INSERT INTO ' <<
      table_alias <<
      ' ' <<
      sql_columns <<
      ' VALUES(' <<
      ('?' * arguments.size).scan(/./).join(',') <<
      ')'

    Log.instance.log('INSERT: ' << sql_insert)
    Log.instance.log('ARGS:   ' << arguments.join(','))
    statement = @db_handle.prepare(sql_insert)
    statement.execute(*arguments)
    statement.finish
  end

  def pg__update(options, table_alias, update_columns, update_arguments, query, arguments)
    unless update_columns.size == update_arguments.size then
      raise SessionError.new('Number of columns and number of arguments mismatch.')
    end
    sql_update = 'UPDATE ONLY ' << table_alias << ' SET '
    sql_update_columns = []
    update_columns.each_index { |i|
      sql_update_columns << "#{update_columns[i]}=?"
    }
    sql_update << sql_update_columns.join(',') << ' WHERE ' << query

    Log.instance.log('SQL UPDATE STMT: ' << sql_update)
    Log.instance.log('SQL UPDATE ARGS: ' << (update_arguments + arguments) * ',')
    statement = @db_handle.prepare(sql_update)
    statement.execute(*(update_arguments + arguments))
    statement.finish
  end

  # MySQL implementation:

  def mysql__create_gazebo_tables()
    # TODO Implement Gazebo database creation.
  end

  def mysql__get_databases()
    return @db_handle.execute('SHOW DATABASES')
  end

  def mysql__get_tables()
    return @db_handle.execute('SHOW TABLES')
  end
end

class SessionError < RuntimeError

end

class GazeboAuthenticationError < RuntimeError

end
