#   Copyright and License: see LICENSE file
#
#   Contributors:
#	   * Joachim Baran

class InterpreterJSON

  def execute(session_id, request)
    begin
      internal_execute(session_id, request)
    rescue DBI::Error
      response = {
        'id' => @request['id'],
        'error' => {
          'origin' => 2,
          'code' => 0,
          'message' => 'Database error'
        },
        'result' => nil
      }
      return { 'session' => session_id, 'response' => JSON.generate(response) }
    end
  end

  private

  def internal_execute(session_id, request)
    @request = JSON.parse(request)

    ret_val = nil
    case @request['method']

    # Connecting/disconnection
    #
    when 'connect' then
      ret_val = connect(session_id)
    when 'disconnect' then
      ret_val = disconnect(session_id)

    # Session status
    #
    when 'status' then
      ret_val = status(session_id)

    # SQL queries
    #
    when 'query' then
      ret_val = query(session_id)
    when 'query_union' then
      ret_val = query_union(session_id)
    when 'query_data' then
      ret_val = query_data(session_id)

    # User-/data-creation and modification
    #
    when 'create_data' then
      ret_val = create_data(session_id)
    when 'create_user' then
      ret_val = internal_create(session_id, 'gazebo_users',
        [
          'username', 'firstname', 'lastname', 'email',
          'deactivated', 'create_user', 'create_group', 'permissions_user'
        ]
      )
    when 'create_group' then
      ret_val = internal_create(session_id, 'gazebo_groups',
        [
          'name', 'contact', 'description', 'contribute_permissions',
          'visible_permissions', 'modify_permissions'
        ]
      )
    when 'update_data' then
      ret_val = update_data(session_id)
    when 'update_user' then
      ret_val = internal_update(session_id, 'gazebo_users', 'username',
        [
          'firstname', 'lastname', 'email', 'deactivated', 'create_user',
          'create_group', 'permissions_user'
        ]
      )
    when 'update_group' then
      ret_val = internal_update(session_id, 'gazebo_groups', 'name',
        [
          'contact', 'description', 'contribute_permissions',
          'visible_permissions', 'modify_permissions'
        ]
      )
    when 'delete_group' then
      ret_val = internal_delete(session_id, 'gazebo_groups g', 'g.name = ?', [ 'gazebo_metadata m', 'gazebo_subscriptions s' ], 'g.group_id = m.group_id OR g.group_id = s.group_id')

    when 'create_metadata' then
      ret_val = internal_create(session_id, 'gazebo_metadata',
        [
          'data_id', 'group_id'
        ]
      )
    when 'delete_metadata' then
      ret_val = internal_delete(session_id, 'gazebo_metadata m', 'm.data_id = ?', nil, nil)
    when 'delete_data' then
      ret_val = internal_delete(session_id, 'gazebo_metadata m', 'm.data_id = ?', [ 'gazebo_groups g' ], 'g.group_id = m.group_id')

    # User-/group-lists
    #
    when 'get_userlist' then
      ret_val = get_list(session_id, [ 'u.username' ], [ 'gazebo_users u', 'gazebo_subscriptions s' ], nil, [], true)
    when 'get_grouplist' then
      ret_val = get_list(session_id, [ 'g.name' ], [ 'gazebo_groups g', 'gazebo_subscriptions s' ], nil, [ 'g.group_id', 'g.description', 'g.contribute_permissions', 'g.visible_permissions', 'g.modify_permissions' ])

    # Details about users and groups
    #
    when 'get_userdetails' then
      ret_val = get_details(
        session_id,
        [
          'username',  'created_by', 'created_on', 'firstname', 'lastname',
          'email', 'deactivated', 'create_user', 'create_group', 'permissions_user'
        ],
        'gazebo_users'
      )
    when 'get_groupdetails' then
      ret_val = get_details(
        session_id,
        [
          'name', 'group_id', 'created_by', 'created_on', 'contact',
          'description', 'contribute_permissions', 'visible_permissions', 'modify_permissions'
        ],
        'gazebo_groups'
      )

    # Subscription information and management
    #
    when 'get_usersubscriptions' then
      ret_val = get_list(session_id,
        [ 'g.name', 'g.created_by', 's.group_admin' ],
        [ 'gazebo_groups g', 'gazebo_subscriptions s' ],
        nil,
        []
      )
    when 'get_groupsubscriptions' then
      ret_val = get_list(session_id,
        [ 'u.username', 's.group_admin' ],
        [ 'gazebo_users u', 'gazebo_subscriptions s' ],
        'gazebo_subscriptions s',
        []
      )
    when 'update_subscriptions' then
      ret_val = internal_subscription_update(session_id)
    end

    log(session_id, request)

    return ret_val
  end

  def connect(session_id)
    @session = Session.new('Pg', session_id)
    (options, host, port, username, password) = @request['params']
    session_id = @session.connect(options, nil, username, password)
    response = { 'id' => @request['id'], 'error' => nil, 'result' => true }
    return { 'session' => session_id, 'response' => JSON.generate(response) }
  rescue GazeboAuthenticationError => e
    response = { 'id' => @request['id'], 'error' => e.to_s, 'result' => false }
    return { 'session' => '-', 'response' => JSON.generate(response) }
  end

  def disconnect(session_id)
    @session = Session.new('Pg', session_id)
    session_id = @session.connect({}, nil, 'gazebo', 'gazebo')
    @session.disconnect()
    response = { 'id' => @request['id'], 'error' => nil, 'result' => true }
    return { 'session' => '-', 'response' => JSON.generate(response) }
  rescue GazeboAuthenticationError => e
    response = { 'id' => @request['id'], 'error' => e.to_s, 'result' => false }
    return { 'session' => '-', 'response' => JSON.generate(response) }
  end

  def log(session_id, json_request)
    log_session = Session.new('Pg', session_id)
    log_session.connect({ 'restricted' => true }, nil, 'gazebo', 'gazebo')
    log_session.log(json_request)
  end

  def status(session_id)
    @session = Session.new('Pg', session_id)
    @session.connect({ 'restricted' => true }, nil, 'gazebo', 'gazebo')
    status = @session.get_status()
    response = { 'id' => @request['id'], 'error' => nil, 'result' => status }
    session_id = '-' unless status['logged_in']
    return { 'session' => session_id, 'response' => JSON.generate(response) }
  end

  def query(session_id)
    @session = Session.new('Pg', session_id)
    (options, database, result_set_columns, table_aliases, query, arguments) = @request['params']
    warn 'Query: ' << database << ' Query: ' << query
    @session.connect({}, database, 'gazebo', 'gazebo')
    result_set = @session.query(options, result_set_columns, table_aliases, query, arguments)
    warn 'Result: ' << result_set.to_s
    warn 'Is null: ' << (result_set == nil).to_s
    response = { 'id' => @request['id'], 'error' => nil, 'result' => result_set }
    return { 'session' => session_id, 'response' => JSON.generate(response) }
  end

  def query_union(session_id)
    @session = Session.new('Pg', session_id)
    (options, database, result_set_columns, table_aliases, queries, arguments) = @request['params']
    @session.connect({}, database, 'gazebo', 'gazebo')
    result_set = @session.query_union(options, result_set_columns, table_aliases, queries, arguments)
    response = { 'id' => @request['id'], 'error' => nil, 'result' => result_set }
    return { 'session' => session_id, 'response' => JSON.generate(response) }
  end

  def query_data(session_id)
    @session = Session.new('Pg', session_id)
    (options, database, result_set_columns, table_aliases, query, arguments) = @request['params']
    warn 'Data Query: ' << database << ' (NOT!) Query: ' << query
    @session.connect({}, 'gazebo', nil, nil)
    warn 'Connected...'
    result_set = @session.query_data(options, result_set_columns, table_aliases, query, arguments)
    warn 'Result Set: ' << result_set.to_s
    response = { 'id' => @request['id'], 'error' => nil, 'result' => result_set }
    return { 'session' => session_id, 'response' => JSON.generate(response) }
  end

  def get_list(session_id, columns, tables, join_table, detailed_columns, userlist=nil)
    @session = Session.new('Pg', session_id)
    (options, query, arguments) = @request['params']
    @session.connect({ 'restricted' => true }, nil, 'gazebo', 'gazebo')
    columns += detailed_columns if options['detailed']
    # TODO Check where 'comprehensive' is applicable..
    if options['queryDataID'] then
      options['comprehensive'] = true
      columns << 'm1.data_id'
      join_table = 'gazebo_metadata m1'
      query << ' AND m1.group_id = g.group_id AND m1.data_id = ?'
      arguments << options['queryDataID']
    end
    # TODO 'contributeOnly' should only be applicable to 'get_grouplist'
    permission_column = nil
    permission_column = 'visible_permissions' unless userlist
    permission_column = 'contribute_permissions' if options['contributeOnly']
    if permission_column then
      tables << 'gazebo_groups g0'
      query << ' AND (' <<
        # Public groups:
        "g0.#{permission_column} = #{ONLY_ADMINS} OR " <<
        # Subscribed groups:
        '(' <<
          "g0.#{permission_column} = #{ADMINS_AND_SUBS} AND " <<
          'g0.group_id IN (' <<
            'SELECT group_id FROM gazebo_subscriptions WHERE username = ?' <<
          ')' <<
        ') OR ' <<
        # Administered groups:
        '(' <<
          "g0.#{permission_column} = #{EVERYONE} AND " <<
          'g0.group_id IN (' <<
            'SELECT group_id FROM gazebo_subscriptions WHERE ' <<
            'username = ? AND group_admin = true' <<
          ')' <<
        ') OR ' <<
        # Group creator:
        'g0.created_by = ?' <<
      ')'
      username = @session.get_username_by_session
      arguments << username << username << username
    end
    #
    result_set = @session.get_list(columns, tables, join_table, options, query, arguments)
    response = { 'id' => @request['id'], 'error' => nil, 'result' => result_set }
    return { 'session' => session_id, 'response' => JSON.generate(response) }
  end

  def get_details(session_id, columns, table)
    @session = Session.new('Pg', session_id)
    (options, argument) = @request['params']
    @session.connect({ 'restricted' => true }, nil, 'gazebo', 'gazebo')
    result_set = @session.get_list(columns, [ table ], nil, options, "#{columns[0]} = ?", [ argument ])
    response = { 'id' => @request['id'], 'error' => nil, 'result' => result_set }
    return { 'session' => session_id, 'response' => JSON.generate(response) }
  end

  def create_data(session_id)
    @session = Session.new('Pg', session_id)
    (options, database, table_alias, columns, arguments) = @request['params']
    @session.connect({}, 'gazebo', nil, nil)
    result_set = @session.create_data(options, table_alias, columns, arguments)
    response = { 'id' => @request['id'], 'error' => nil, 'result' => result_set }
    return { 'session' => session_id, 'response' => JSON.generate(response) }
  end

  def update_data(session_id)
    @session = Session.new('Pg', session_id)
    (options, database, table_alias, data_id, columns, arguments) = @request['params']
    @session.connect({}, 'gazebo', nil, nil)
    result = @session.update_data(options, data_id, table_alias, columns, arguments)
    response = { 'id' => @request['id'], 'error' => nil, 'result' => result }
    return { 'session' => session_id, 'response' => JSON.generate(response) }
  end

  def internal_create(session_id, table, columns)
    @session = Session.new('Pg', session_id)
    (options, arguments) = @request['params']
    @session.connect({ 'restricted' => true }, nil, 'gazebo', 'gazebo')
    status = @session.create_internals(options, table, columns, arguments)
    response = { 'id' => @request['id'], 'error' => nil, 'result' => status }
    return { 'session' => session_id, 'response' => JSON.generate(response) }
  end

  def internal_delete(session_id, table, query, constraint_tables, constraint_query)
    @session = Session.new('Pg', session_id)
    (options, arguments) = @request['params']
    @session.connect({ 'restricted' => true }, nil, 'gazebo', 'gazebo')

    if constraint_tables and constraint_query then
      # Check constraints: if constraint returns a non-empty result, then abort...
      constraint_result = @session.simple_query([ 'COUNT(*)' ], [ table ] | constraint_tables, query + " AND (#{constraint_query})", arguments)
      unless constraint_result[0][0] == 0 then
        response = {
          'id' => @request['id'],
          'error' => {
            'origin' => 2,
            'code' => 0,
            'message' => 'Resource is still linked to other sources.'
          },
          'result' => nil
        }
        return { 'session' => session_id, 'response' => JSON.generate(response) }
      end
    end

    status = @session.delete_internals(options, table, query, arguments)
    response = { 'id' => @request['id'], 'error' => nil, 'result' => status }
    return { 'session' => session_id, 'response' => JSON.generate(response) }
  end

  def internal_update(session_id, table, key_column, update_columns)
    @session = Session.new('Pg', session_id)
    (options, key, update_arguments) = @request['params']
    @session.connect({ 'restricted' => true }, nil, 'gazebo', 'gazebo')
    status = @session.update_internals(options, table, key, key_column, update_columns, update_arguments)
    response = { 'id' => @request['id'], 'error' => nil, 'result' => status }
    return { 'session' => session_id, 'response' => JSON.generate(response) }
  end

  def internal_subscription_update(session_id)
    @session = Session.new('Pg', session_id)
    (options, name, subscription_list) = @request['params']
    @session.connect({ 'restricted' => true }, nil, 'gazebo', 'gazebo')
    warn 'GO!'
    subscription_list.each { |subscription|
      ( username, is_admin, is_subscribed ) = subscription
      warn '--> ' << subscription.join(';')
      @session.update_subscription(options, name, username, is_admin, is_subscribed)
    }
    warn 'SUBSCRIPTION UPDATED!'
    # TODO Return true for now.
    response = { 'id' => @request['id'], 'error' => nil, 'result' => true }
    return { 'session' => session_id, 'response' => JSON.generate(response) }
  end
end
